!---------------------------------------------------------------------------------------
! PREP_REGIONS: mod_flexpart
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> mod_flexpart
!!
!! Purpose:    Module for interfacing with flexpart output.
!!
!---------------------------------------------------------------------------------------


module mod_flexpart

  use mod_var
  use mod_dates
  use mod_settings

  implicit none
  private

  public :: read_header, read_grid, read_init, read_factor, &
            read_grid_std, read_init_std, get_nread, read_dates

  contains

  ! --------------------------------------------------
  ! read_header
  ! --------------------------------------------------
  !> read_header
  !! 
  !! Purpose:  Reads the flexpart header file and
  !!           initiates general flexpart variables.
  !!
  ! --------------------------------------------------

  subroutine read_header(filename, numpoint, ibdate, ibtime, releases, &
                           numx, numy, numz, bndx, bndy, delx, dely)

    use mod_var

    implicit none

    character(len=max_path_len),  intent(in)     :: filename
    integer,                      intent(in out) :: numpoint
    integer,                      intent(in out) :: ibdate, ibtime   
    integer, dimension(maxpoint), intent(in out) :: releases           
    integer,                      intent(in out) :: numx, numy, numz
    real,                         intent(in out) :: bndx, bndy, delx, dely

    character(len=29)                            :: flexversion
    logical                                      :: lexist
    integer                                      :: nageclass, imethod, nspec
    integer                                      :: loutstep, loutaver, loutsample
    integer                                      :: jjjjmmdd, ihmmss
    real                                         :: xp2, yp2
    character(len=10), dimension(maxspec)        :: species
    integer, dimension(:), allocatable           :: ireleasestart, ireleaseend
    integer, dimension(:), allocatable           :: npart, nkind
    character(len=45), dimension(:), allocatable :: compoint
    real, dimension(:), allocatable              :: xpoint, ypoint, zpoint1, zpoint2
    real, dimension(:,:), allocatable            :: xmass
    integer, dimension(10)                       :: lage
    integer                                      :: n, i, j, ierr, dummy
 
    inquire ( file=trim(filename),exist=lexist )
    if ( .not.lexist ) then
      write(*,*) 'ERROR: cannot find: '//trim(filename)
      stop
    endif
    open(100,file=filename,form='unformatted',action='read',status='old',iostat=ierr)
    if(ierr.gt.0) then
      write(*,*) 'ERROR: cannot read: '//trim(filename)
      stop
    endif

    read(100) ibdate, ibtime, flexversion
    read(100) loutstep, loutaver, loutsample
    read(100) bndx, bndy, numx, numy, delx, dely
    read(100) numz, (outheight(i), i = 1, numz)
    read(100) jjjjmmdd, ihmmss
    read(100) nspec, numz
    nspec=nspec/3
    do n=1,nspec
      read(100) dummy, species(n)
      read(100) dummy, species(n)
      read(100) numz, species(n)
    end do
    read(100) numpoint

    allocate( ireleasestart(numpoint) )
    allocate( ireleaseend(numpoint) )
    allocate( xpoint(numpoint) )
    allocate( ypoint(numpoint) )
    allocate( zpoint1(numpoint) )
    allocate( zpoint2(numpoint) )
    allocate( npart(numpoint) )
    allocate( nkind(numpoint) )
    allocate( xmass(numpoint,nspec) )
    allocate( compoint(numpoint) )

    do i = 1, numpoint
      read(100) ireleasestart(i), ireleaseend(i)
      read(100) xpoint(i), ypoint(i), xp2, yp2, zpoint1(i), zpoint2(i)
      read(100) npart(i), nkind(i)
      read(100) compoint(i)
       do j=1,nspec
         read(100)
         read(100)
         read(100) xmass(i,j)
       end do
    end do
    read(100) imethod
    read(100) nageclass,(lage(i), i = 1, nageclass)
    close(100)

    releases(1:numpoint) = ireleasestart(:)
    ageclass = lage(1)
    trajdays = ageclass/3600/24
    ndgrid = abs(24*3600/loutstep)
 
    if( real(numx)*delx.eq.360. ) then
      ! global domain - check offset relative to 180W
      nxshift = int((bndx + 180.)/delx) 
      bndx = -180.
    else 
      ! not global domain
      nxshift = 0
    endif

    deallocate(ireleasestart)
    deallocate(ireleaseend)
    deallocate(xpoint)
    deallocate(ypoint)
    deallocate(zpoint1)
    deallocate(zpoint2)
    deallocate(npart)
    deallocate(nkind)
    deallocate(xmass)
    deallocate(compoint)

  end subroutine read_header

  ! --------------------------------------------------
  ! read_grid
  ! --------------------------------------------------
  !> read_grid
  !!
  !! Purpose:  Reads the grid_time files where there
  !!           is one grid_time file per release.
  !!
  ! --------------------------------------------------

  subroutine read_grid(config, filename, filedates, jd, numx, numy, grid, gtime)

    use mod_var

    implicit none

    type (config_t),                  intent(in)     :: config
    character(len=max_path_len),      intent(in)     :: filename, filedates
    real(kind=8),                     intent(in)     :: jd
    integer,                          intent(in)     :: numx, numy
    real, dimension(numx,numy,ngrid), intent(in out) :: grid
    real(kind=8), dimension(ngrid),   intent(in out) :: gtime

    real, parameter                                  :: scaleconc=1.e12
    real, parameter                                  :: smallnum=tiny(0.)
    logical                                          :: lexist
    integer                                          :: ierr
    integer                                          :: nread
    integer                                          :: sp_count_i, sp_count_r
    integer                                          :: jjjjmmdd, hhmiss
    integer                                          :: ii, ir, fact, n, nt, jy, ix, ind
    integer, dimension(numx*numy)                    :: sparse_dump_i
    real, dimension(numx*numy)                       :: sparse_dump_r
    real, dimension(:,:,:), allocatable              :: work
    integer(kind=4), dimension(:), allocatable       :: times
    real(kind=8), dimension(:), allocatable          :: dates
    real(kind=8)                                     :: jdrel
    real(kind=8), dimension(:), allocatable          :: jdtime

    ! read footprint dates
    inquire ( file=trim(filedates),exist=lexist )
    if ( .not.lexist ) then
      write(*,*) 'ERROR: cannot find '//trim(filedates)
      stop
    endif
    call get_nread(filedates, nread)
    allocate( dates(nread) )
    call read_dates(filedates, nread, dates)

    ! initialize
    allocate( work(numx,numy,ngrid) )
    allocate( times(ngrid) )
    allocate( jdtime(ngrid) )
    work(:,:,:) = 0.
    grid(:,:,:) = 0.
    times(:) = 0.
    gtime(:) = 0d0

    ! open grid_time file
    inquire ( file=trim(filename),exist=lexist )
    print*, filename
    if ( .not.lexist ) then
      write(*,*) 'WARNING: cannot find '//trim(filename)
      go to 10
    endif
    open(100,file=trim(filename),form='unformatted',action='read',status='old',iostat=ierr)
    if( ierr.ne.0 ) then
      write(*,*) 'WARNING: cannot read '//trim(filename)
      go to 10
    endif

    ! read footprints
    n = 0
    do nt = 1, ngrid
      read(100,iostat=ierr,end=20) jjjjmmdd
      read(100,iostat=ierr,end=20) hhmiss
      ! note: assumes loutaverage of at least 1 hour
      ! inversion output format has no wet/dry deposition
      times(nt)=jjjjmmdd*100+hhmiss/10000
      fact = 1
      read(100,iostat=ierr,end=20) sp_count_i
      read(100,iostat=ierr,end=20) (sparse_dump_i(ix), ix=1,sp_count_i)
      read(100,iostat=ierr,end=20) sp_count_r
      read(100,iostat=ierr,end=20) (sparse_dump_r(ix), ix=1,sp_count_r)
      ii = 0
      do ir = 1, sp_count_r
        if ((sparse_dump_r(ir)*fact).gt.smallnum) then
          ii = ii + 1
          n = sparse_dump_i(ii)
          fact = fact*(-1.)
        else
          n = n + 1
        endif
        jy = (n - numx * numy)/numx
        ix = n - numx * numy - numx * jy
        work(ix+1,jy+1,nt) = abs(sparse_dump_r(ir)) * scaleconc
      end do 
    end do 
20  continue
    if ( ierr.ne.0 ) write(*,fmt='(A29,I2)') 'WARNING: read_grid stopped at nt ',nt

    close(100)

    ! shift grid by nxshift along longitudinal dimension
    work = cshift(work, shift=-1*nxshift, dim=1)

    ! calculate footprint time in julian days
    do nt = 1, ngrid
      hhmiss = (times(nt) - (times(nt)/100)*100)*10000
      jdtime(nt) = juldate(times(nt)/100,hhmiss)
    end do

    ! index to footprints for current release
    ! jd is julian day of release
    ! ndgrid is number of grids per day
    jdrel = jd - 1./real(ndgrid)
    
    ! reverse order in time dimension
    do nt = 1, ngrid
      n = ngrid - nt + 1
      grid(:,:,n) = work(:,:,nt)
      gtime(n) = jdtime(nt)
    end do

10  continue

    deallocate( work )
    deallocate( times )
    deallocate( dates )
    deallocate( jdtime )

  end subroutine read_grid

  ! --------------------------------------------------
  ! read_grid_std
  ! --------------------------------------------------
  !> read_grid_std
  !!
  !! Purpose:  Reads the grid_time files where there
  !!           is one grid_time file per footprint
  !!           with time dimension of releases
  !!
  ! --------------------------------------------------

  subroutine read_grid_std(config, filename, jd, numx, numy, grid, gtime)

    use mod_var

    implicit none

    type (config_t),                  intent(in)     :: config
    character(len=max_path_len),      intent(in)     :: filename
    real(kind=8),                     intent(in)     :: jd
    integer,                          intent(in)     :: numx, numy
    real, dimension(numx,numy,ngrid), intent(in out) :: grid
    real(kind=8), dimension(ngrid),   intent(in out) :: gtime

    real, parameter                                  :: scaleconc=1.e12
    real, parameter                                  :: smallnum=tiny(0.)
    logical                                          :: lexist
    integer                                          :: ierr
    integer                                          :: sp_count_i, sp_count_r
    integer                                          :: jjjjmmdd, hhmiss, itime
    integer                                          :: ii, ir, fact, n
    integer                                          :: kp, ks, kz, ix, jy
    integer, dimension(numx*numy)                    :: sparse_dump_i
    real, dimension(numx*numy)                       :: sparse_dump_r
    real, dimension(numx*numy)                       :: sparse_dump_u

    grid(:,:,:) = 0.
    gtime(:) = 0d0

    ! open grid_time file
    inquire ( file=trim(filename),exist=lexist )
    if ( .not.lexist ) then
      write(*,*) 'WARNING: cannot find '//trim(filename)
      go to 10
    endif
    open(100,file=trim(filename),form='unformatted',action='read',status='old',iostat=ierr)
    if( ierr.ne.0 ) then
      write(*,*) 'WARNING: cannot read '//trim(filename)
      go to 10
    endif

    ! grid time equal to release time
    gtime = jd + reltime/3600d0/24d0

    read(100) itime
    do ks=1,1
      do kp=1,ngrid
        ! read wet deposition
        read(100,iostat=ierr,end=20) sp_count_i
        read(100,iostat=ierr,end=20) (sparse_dump_i(ix),ix=1,sp_count_i)
        read(100,iostat=ierr,end=20) sp_count_r
        read(100,iostat=ierr,end=20) (sparse_dump_r(ix),ix=1,sp_count_r)
!        if ( config%lnested ) then
!          read(100) sp_count_r
!          read(100) (sparse_dump_u(ix),ix=1,sp_count_r)
!        endif
        ! read dry deposition
        read(100,iostat=ierr,end=20) sp_count_i
        read(100,iostat=ierr,end=20) (sparse_dump_i(ix),ix=1,sp_count_i)
        read(100,iostat=ierr,end=20) sp_count_r
        read(100,iostat=ierr,end=20) (sparse_dump_r(ix),ix=1,sp_count_r)
!        if ( config%lnested ) then
!          read(100) sp_count_r
!          read(100) (sparse_dump_u(ix),ix=1,sp_count_r)
!        endif
        ! read concentrations
        fact=1
        read(100,iostat=ierr,end=20) sp_count_i
        read(100,iostat=ierr,end=20) (sparse_dump_i(ix),ix=1,sp_count_i)
        read(100,iostat=ierr,end=20) sp_count_r
        read(100,iostat=ierr,end=20) (sparse_dump_r(ix),ix=1,sp_count_r)
        ii=0
        do ir=1,sp_count_r
          if ((sparse_dump_r(ir)*fact).gt.smallnum) then
            ii=ii+1
            n=sparse_dump_i(ii)
            fact=fact*(-1.)
          else
            n=n+1
          endif
          kz=n/(numx*numy)
          jy=(n-kz*numx*numy)/numx
          ix=n-numx*numy*kz-numx*jy
          grid(ix+1,jy+1,kp)=abs(sparse_dump_r(ir))*scaleconc
        end do ! sp_count_r
      end do ! ngrid 
    end do ! nspec
20  continue
    if ( ierr.ne.0 ) write(*,fmt='(A29,I2)') 'WARNING: read_grid stopped at kp ',kp
    close(100)

    ! shift grid by nxshift along longitudinal dimension
    grid = cshift(grid, shift=-1*nxshift, dim=1)    

10  continue

  end subroutine read_grid_std

  ! --------------------------------------------------
  ! read_init
  ! --------------------------------------------------
  !> read_init
  !! 
  !! Purpose:  Reads the grid_initial files where 
  !!           there is one file per release.
  !!
  ! --------------------------------------------------

  subroutine read_init(filename, gridinit)

    use mod_var

    implicit none

    character(len=max_path_len),           intent(in)     :: filename
    real, dimension(nxgrid,nygrid,nzgrid), intent(in out) :: gridinit

    real, parameter                                       :: scaleconc=1.e12
    real, parameter                                       :: smallnum=tiny(0.)
    logical                                               :: lexist
    integer                                               :: ierr
    integer                                               :: sp_count_i, sp_count_r
    integer                                               :: jjjjmmdd, hhmiss
    integer                                               :: ii, ir, n, nt, jy, ix, kz
    real                                                  :: fact
    integer, dimension(nxgrid*nygrid*nzgrid)              :: sparse_dump_i
    real, dimension(nxgrid*nygrid*nzgrid)                 :: sparse_dump_r
    
    gridinit(:,:,:)=0.

    inquire ( file=trim(filename),exist=lexist )
    if ( .not.lexist ) then
      write(*,*) 'WARNING: cannot find '//trim(filename)
      go to 10
    endif
    open(100, file=trim(filename), form='unformatted', action='read', status='old', iostat=ierr)
    if ( ierr.ne.0 ) then
      write(*,*) 'WARNING: cannot read '//trim(filename)
      go to 10
    endif

    read(100) jjjjmmdd
    read(100) hhmiss

    fact = 1.
    read(100) sp_count_i
    read(100) (sparse_dump_i(ix), ix=1, sp_count_i)
    read(100) sp_count_r
    read(100) (sparse_dump_r(ix), ix=1, sp_count_r)
    ii = 0
    do ir = 1, sp_count_r
      if ((sparse_dump_r(ir)*fact).gt.smallnum) then
        ii = ii + 1
        n = sparse_dump_i(ii)
        fact = fact*(-1.)
      else
        n = n + 1
      endif
      kz = n/(nxgrid*nygrid)
      jy = (n - kz*nxgrid*nygrid)/nxgrid
      ix = n - nxgrid*nygrid*kz-nxgrid*jy
      gridinit(ix+1,jy+1,kz) = abs(sparse_dump_r(ir))*scaleconc
    end do 

    ! shift grid by xshift along longitudinal dimension
    gridinit = cshift(gridinit, shift=-1*nxshift, dim=1)

    close(100)

10  continue

  end subroutine read_init

  ! --------------------------------------------------
  ! read_init_std
  ! --------------------------------------------------
  !> read_init_std
  !! 
  !! Purpose:  Reads the grid_initial files where 
  !!           there is one file with time dimension 
  !!           of releases.
  !!
  ! --------------------------------------------------

  subroutine read_init_std(filename, gridinit)

    use mod_var

    implicit none

    character(len=max_path_len),           intent(in)     :: filename
    real, dimension(nxgrid,nygrid,nzgrid), intent(in out) :: gridinit

  end subroutine read_init_std

  ! --------------------------------------------------
  ! get_nread
  ! --------------------------------------------------
  !> get_nread
  !!
  !! Purpose:  Gets the number of time steps in the
  !!           grid_time files.
  !!
  ! --------------------------------------------------

  subroutine get_nread(filename, nread)

    use mod_var

    implicit none

    character(len=max_path_len),          intent(in)     :: filename
    integer,                              intent(in out) :: nread

    character(len=14), dimension(1000)                   :: tmp
    logical                                              :: lexist
    integer                                              :: ierr, n

    inquire ( file=trim(filename),exist=lexist )
    if ( .not.lexist ) then
      write(*,*) 'ERROR: cannot find '//trim(filename)
      stop
    endif
    open(100,file=trim(filename),action='read',status='old',iostat=ierr)
    if( ierr.ne.0 ) then
      write(*,*) 'ERROR: cannot read '//trim(filename)
      stop
    endif

    n = 1
    do while (ierr.eq.0)
      read(100,fmt='(A14)',iostat=ierr,end=10) tmp(n)
      n = n + 1
    enddo
10  continue

    close(100)

    nread = n - 1

  end subroutine get_nread

  ! --------------------------------------------------
  ! read_dates
  ! --------------------------------------------------
  !> read_dates
  !! 
  !! Purpose:  Reads the dates for each time step in
  !!           the grid_time files.
  !!
  ! --------------------------------------------------

  subroutine read_dates(filename, nread, dates)

    use mod_var
   
    implicit none

    character(len=max_path_len),          intent(in)     :: filename
    integer,                              intent(in)     :: nread
    real(kind=8), dimension(nread),       intent(in out) :: dates

    character(len=14)                                    :: tmp
    integer                                              :: ierr, n

    open(100,file=filename,action='read',status='old',iostat=ierr)
    if( ierr.ne.0 ) then
      write(*,*) 'ERROR: cannot read: '//trim(filename)
      stop
    endif 

    do n = 1, nread
      read(100,fmt='(A14)',iostat=ierr,end=10) tmp
      read(tmp,*) dates(n)
    enddo
10  continue

    close(100)

  end subroutine read_dates

  ! --------------------------------------------------
  ! read_factor
  ! --------------------------------------------------
  !> read_factor
  !!
  !! Purpose:  Reads the correction factor (i.e. ratio
  !!           of the wet to dry air density for 
  !!           correcting mixing ratios for dry air).
  !!
  ! --------------------------------------------------

  subroutine read_factor(filename, nread, numx, numy, xshift, factor)

    use mod_var

    implicit none

    character(len=max_path_len),          intent(in)     :: filename
    integer,                              intent(in)     :: nread
    integer,                              intent(in)     :: numx, numy, xshift
    real, dimension(numx,numy,nread),     intent(in out) :: factor

    real, parameter                                      :: smallnum=tiny(0.)
    logical                                              :: lexist
    integer                                              :: ierr
    integer                                              :: sp_count_i, sp_count_r
    integer                                              :: ii, ir, n, nt, jy, ix
    real                                                 :: fact
    integer, dimension(numx*numy)                        :: sparse_dump_i
    real, dimension(numx*numy)                           :: sparse_dump_r

    ! open factor file
    inquire ( file=trim(filename),exist=lexist )
    if ( .not.lexist ) then
      write(*,*) 'WARNING: cannot find '//trim(filename)
      go to 10
    endif
    open(100,file=trim(filename),form='unformatted',action='read',status='old',iostat=ierr)
    if( ierr.ne.0 ) then
      write(*,*) 'WARNING: cannot read '//trim(filename)
      go to 10
    endif

    ! read footprints
    do nt = 1, nread
      fact = 1.
      read(100) sp_count_i
      read(100) (sparse_dump_i(ix), ix=1,sp_count_i)
      read(100) sp_count_r
      read(100) (sparse_dump_r(ix), ix=1,sp_count_r)
      ii = 0
      do ir = 1, sp_count_r
        if ((sparse_dump_r(ir)*fact).gt.smallnum) then
          ii = ii + 1
          n = sparse_dump_i(ii)
          fact = fact*(-1.)
        else
          n = n + 1
        endif
        jy = (n - numx * numy)/numx
        ix = n - numx * numy - numx * jy
        factor(ix+1,jy+1,nt) = abs(sparse_dump_r(ir))
      end do
    end do

    close(100)

    ! shift grid by xshift along longitudinal dimension
    factor = cshift(factor, shift=-1*xshift, dim=1)

    10 continue

  end subroutine read_factor

  ! --------------------------------------------------

end module mod_flexpart


