!---------------------------------------------------------------------------------------
! FLEXINVERT: calc_conc 
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> calc_conc
!!
!! Purpose:    Calculates fixed mixing ratios.
!!
!! Details:    Calculates the fixed components of the mixing ratios
!!
!! Interface:
!!
!!    Inputs
!!             config  -  configuration settings data structure
!!             fluxes  -  flux data structure
!!             obs     -  observation data structure
!!             ngrid   -  time dimension of hnest
!!             hnest   -  transport operator nested domain
!!             ind     -  index to low temporal resolution fluxes
!!             iobs    -  index to obs   
!!
!---------------------------------------------------------------------------------------

subroutine calc_conc(config, fluxes, obs, ngrid, gtime, hnest, hbkg, iobs, ix1, ix2, jy1, jy2)

  use mod_settings
  use mod_var
  use mod_dates
  use mod_fluxes
  use mod_obs

  implicit none

  type (config_t),                          intent (in)     :: config
  type (fluxes_t),                          intent (in)     :: fluxes
  type (obs_t),                             intent (in out) :: obs
  real(kind=8), dimension(maxngrid),        intent (in)     :: gtime 
  real, dimension(nxregrid,nyregrid,ngrid), intent (in)     :: hnest
  real, dimension(nxgrid,nygrid),           intent (in)     :: hbkg
  integer,                                  intent (in)     :: ngrid, iobs, ix1, ix2, jy1, jy2

  integer                                                   :: n, ns, ni, flag, ind, ilo, ihi, ierr
  real, dimension(nxgrid,nygrid)                            :: flxbg
  real                                                      :: bkgerr, ffferr

  ! background
  ! ----------

  ! index to global fluxes for start of footprint time
  flag = 0
  ni = 0
  do while ( (flag.eq.0).and.(ni.lt.ntstate) )
    ni = ni + 1
    if ( gtime(1).lt.(fluxes%time(ni)+real(statres,kind=8)) ) flag = 1
  enddo
  ind = min(ni, ntstate)

  ! background fluxes
  if ( trim(config%spec).eq.'co2' ) then
    ! CO2 species
    flxbg = fluxes%flxnee(:,:,ind) + &
                    fluxes%flxff(:,:,ind) + &
                    fluxes%flxocn(:,:,ind) 
  else if (( trim(config%spec).eq.'ghg' ) .or. ( trim(config%spec).eq.'aero' )) then
    ! GHG or aerosol species
    flxbg = fluxes%flx(:,:,ind)
  endif

  ! background mixing ratio 
  obs%bkg(iobs) = sum( hbkg(:,:) * flxbg(:,:) )

  if (iobs.lt.10) print*, 'calc_conc: obs%bkg(iobs) = ',obs%bkg(iobs)

  ! background mixing ratio uncertainty
  bkgerr = ( sum( hbkg(:,:) * abs(flxbg(:,:)) ) * config%flxerr )**2

  ! fixed contributions
  ! -------------------

  if ( trim(config%spec).eq.'co2' ) then
    ! CO2 species
    obs%nee(iobs) = 0.
    obs%fff(iobs) = 0.
    obs%ocn(iobs) = 0.
    ffferr = 0. 
  else
    ! GHG species
    obs%ghg(iobs) = 0.
    ! ocean contribution only if lognormal is true and inc_ocean is false
    obs%ocn(iobs) = 0.
  endif

  do n = 1, ngrid
    ! calculate indice to fluxes
    flag = 0
    ind = 0
    do while ( (flag.eq.0).and.(ind.lt.ntstate) )
      ind = ind + 1
      if ( gtime(n).lt.(fluxes%time(ind)+real(statres,kind=8)) ) flag = 1
    enddo
    ! index to lo-res fluxes 
    ilo = min(ind, ntstate)
    if ( trim(config%spec).eq.'co2' ) then
      ! CO2 species
      ! index to hi-res fluxes (NEE and FFF only)
      ihi = minloc(fluxes%timefp,dim=1,mask=abs(fluxes%timefp-gtime(n)).eq.minval(abs(fluxes%timefp-gtime(n))))
      ! change in mixing ratio
      obs%nee(iobs) = obs%nee(iobs) + sum(hnest(:,:,n) * fluxes%flxnee_nest(:,:,ihi))
      obs%fff(iobs) = obs%fff(iobs) + sum(hnest(:,:,n) * fluxes%flxff_nest(:,:,ihi))
      ffferr = ffferr + (sum(hnest(:,:,n) * fluxes%flxff_nest(:,:,ihi) * config%ffferr))**2
      if ( config%nested ) then
        obs%ocn(iobs) = obs%ocn(iobs) + sum(hnest(:,:,n) * fluxes%flxocn_nest(:,:,ilo))
      else
        obs%ocn(iobs) = obs%ocn(iobs) + sum(hnest(:,:,n) * fluxes%flxocn(ix1:ix2,jy1:jy2,ilo))
      endif
    else
      ! GHG and aerosol species
      if ( config%nested ) then
        ! nested domain
        obs%ghg(iobs) = obs%ghg(iobs) + sum(hnest(:,:,n) * fluxes%flx_nest(:,:,ilo))
      else
        ! no nested domain
        obs%ghg(iobs) = obs%ghg(iobs) + sum(hnest(:,:,n) * fluxes%flx(ix1:ix2,jy1:jy2,ilo))
      endif
      if ( config%lognormal.and..not.config%inc_ocean ) then
        ! since for lognormal scalars of fluxes are optimized, if ocean fluxes are not optimized need
        ! to still account for influence on mixing ratios
        obs%ocn(iobs) = obs%ocn(iobs) + sum(hnest(:,:,n) * fluxes%flxocn(:,:,ilo))
      endif
    endif
  end do ! ngrid

  ! observation space error
  ! -----------------------  

  if ( trim(config%spec).eq.'co2' ) then
    ! CO2
    if ( config%opt_cini ) then
      obs%err(iobs) = sqrt(obs%measerr(iobs)**2 + bkgerr + ffferr)
    else
      obs%err(iobs) = sqrt(obs%measerr(iobs)**2 + (sum(obs%cini(iobs,:))*config%cinierr)**2 + bkgerr + ffferr)
    endif
  else if ( trim(config%spec).eq.'ghg' ) then
    ! GHG 
    if ( config%opt_cini ) then
      obs%err(iobs) = sqrt(obs%measerr(iobs)**2 + bkgerr)
    else
      obs%err(iobs) = sqrt(obs%measerr(iobs)**2 + (sum(obs%cini(iobs,:))*config%cinierr)**2 + bkgerr)
    endif
  else if ( trim(config%spec).eq.'aero' ) then
    ! aerosols
    obs%err(iobs) = sqrt(obs%measerr(iobs)**2 +  bkgerr)
  endif

end subroutine calc_conc


